
/**********************************************************************
 * $Id: estimate.c,v 1.4 1993/01/15 13:04:19 drew Exp $
 **********************************************************************/

/**********************************************************************
 *   Copyright 1990,1991,1992,1993 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <xerion/useful.h>
#include <xerion/simulator.h>
#include <xerion/commands.h>



/**********************************************************************/
typedef struct {
  int	n ;
  Real	sum ;
} ErrorStats ;
/**********************************************************************/
typedef enum {
  UseWeight,
  UseError,
  UseRatio
} UseValue ;
/**********************************************************************/
#define SQR(x)	((x)*(x))
static void	sumStats ARGS((Unit unit, void *data)) ;
/**********************************************************************/
static void 
sumStats(unit, data)
  Unit		unit ;
  void		*data ;
{
  ErrorStats	*stats = (ErrorStats *)data ;
  
  stats->n++ ;
  stats->sum  += SQR(unit->target - unit->output) ;
}
/**********************************************************************/
int 
command_estimateVariance(tokc,tokv)
  int	tokc ;
  char	*tokv[] ;
{
  ExampleSet	exampleSet ;
  ErrorStats	stats ;
  String	name ;
  String	modelName = NULL ;
  UseValue	use	  = UseRatio ;
  Boolean	setScale  = TRUE ;
  Boolean	print     = FALSE ;
  Real		weightVariance, errorVariance, useValue ;
  CostModel	costModel ;
  Group		group ;
  int		oldIdx, idx ;

  IUsage("[-model <model>] [-noSet] [-print] [-weight] [-error] [-ratio]");
  if (GiveHelp(tokc)) {
    ISynopsis("estimate the variance of the network error or weights");
    IHelp
      (IHelpArgs,
       NULL) ;
    return 1 ;
  }

  name = *tokv ;
  for (++tokv, --tokc ; tokc > 0 ; ++tokv, --tokc) {
    if (strncmp(*tokv, "-weight", strlen(*tokv)) == 0) {
      use = UseWeight ;
    } else if (strncmp(*tokv, "-error", strlen(*tokv)) == 0) {
      use = UseError ;
    } else if (strncmp(*tokv, "-ratio", strlen(*tokv)) == 0) {
      use = UseRatio ;
    } else if (strncmp(*tokv, "-noSet", strlen(*tokv)) == 0) {
      setScale = FALSE ;
    } else if (strncmp(*tokv, "-print", strlen(*tokv)) == 0) {
      print = TRUE ;
    } else if (strncmp(*tokv, "-model", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      modelName = *tokv ;
    } else if (*tokv[0] == '-') {
      IErrorAbort("Invalid option: %s", *tokv) ;
    } else {
      break ;
    }
  }

  if (tokc != 0)
    IErrorAbort(IPrintUsage(name, usage)) ;

  if (currentNet == NULL)
    IErrorAbort("No currentNet") ;

  exampleSet = netGetExampleSet(currentNet, TRAINING) ;
  if (exampleSet == NULL)
    IErrorAbort("Unknown example set type: \"0x%0x\".", TRAINING) ;

  if (exampleSet->numExamples == 0)
    IErrorAbort("Example Set is empty") ;

  if (modelName == NULL || strcmp(currentNet->name, modelName) == 0)
    costModel = currentNet->costModel ;
  else if (group = groupFromName(currentNet, modelName))
    costModel = group->costModel ;
  else
    IErrorAbort("Unknown cost model: \"%s\"", modelName) ;

  if (costModel == NULL)
    IErrorAbort("Cost model does not exist!") ;


  if (use != UseWeight) {
    /***********************
     * Now, do the examples
     */
    stats.n	= 0 ;
    stats.sum	= 0.0 ;
    oldIdx = exampleSet->currentIdx ;
    for (idx = 0 ; idx < exampleSet->numExamples ; ++idx) {
      MgetNumber(exampleSet, idx) ;
      MupdateNetActivities(currentNet) ;
      netForAllUnits(currentNet, OUTPUT, sumStats, &stats) ;
    }
    exampleSet->currentIdx = oldIdx ;

    errorVariance = stats.sum/stats.n ;
  }

  if (use != UseError) {
    if (costModel->missingLinks)
      MCMcompact(costModel) ;

    stats.sum = 0.0 ;
    stats.n   = 0 ;
    for (idx = 0 ; idx < costModel->numLinks ; ++idx) {
      stats.sum += SQR(costModel->link[idx]->weight) ;
      stats.n++ ;
    }

    weightVariance = stats.sum/stats.n ;
  }

  switch(use) {
  case UseWeight:
    useValue = weightVariance ;
    break ;
  case UseError:
    useValue = errorVariance ;
    break ;
  case UseRatio:
    useValue = weightVariance/errorVariance ;
    break ;
  default:
    useValue = 0.0 ;
    break ;
  }

  if (print)
    fprintf(dout, "%g\n", useValue) ;
    
  if (setScale)
    MCMsetScale(costModel, useValue ? 1.0/useValue : 0.0) ;

  return 1 ;
}
/**********************************************************************/
